
#if !defined( glwrap_h )
#define glwrap_h

#include "main.h"

#include <GL/GL.h>
#include "myext.h"

#include "textures.h"



//  A color is expressed as a GWColor. This is so that there's 
//  a little bit of type safety involved.
union GWColor {
  GWColor() { }
  GWColor( unsigned int c ) : color( c ) { }
  GWColor( unsigned char a, unsigned char r, unsigned char g, unsigned char b ) : blue( b ), green( g ), red( r ), alpha( a ) { }
//fixme:  this is only true for little-endian platform
  struct {
    unsigned char   blue;
    unsigned char   green;
    unsigned char   red;
    unsigned char   alpha;
  };
  unsigned int      color;

  bool operator==( GWColor const & o ) const
            {
              return color == o.color;
            }
  bool operator!=( GWColor const & o ) const
            {
              return color == o.color;
            }
};

//  A TexFrame is a subsection of a full texture, used to 
//  express an image within a texture when the texture contains 
//  an animation made up of multiple frames.
struct GWTexFrame {
  GWTexFrame( float scaleS, float scaleT, float offsetS, float offsetT )
    : mScaleS( scaleS )
    , mScaleT( scaleT )
    , mOffsetS( offsetS )
    , mOffsetT( offsetT )
    {
    }
  GWTexFrame()
    : mScaleS( 1.f )
    , mScaleT( 1.f )
    , mOffsetS( 0.f )
    , mOffsetT( 0.f )
    {
    }
  bool operator==( GWTexFrame const & o ) const
    {
      return (mScaleS == o.mScaleS) && (mScaleT == o.mScaleT) && 
          (mOffsetS == o.mOffsetS) && (mOffsetT == o.mOffsetT);
    }
  bool operator!=( GWTexFrame const & o ) const
    {
      return !(*this == o );
    }
  void apply() const
    {
      glMatrixMode( GL_TEXTURE );
      if( *this == unity ) {
        glLoadIdentity();
      }
      else {
        static GLfloat f[ 16 ] = {
          0.f, 0.f, 0.f, 0.f,
          0.f, 0.f, 0.f, 0.f,
          0.f, 0.f, 0.f, 0.f,
          0.f, 0.f, 0.f, 1.f,
        }; //  will stay initialized to zeros where we don't use it!
        f[ 0 ] = mScaleS;
        f[ 5 ] = mScaleT;
        f[ 12 ] = mOffsetS;
        f[ 13 ] = mOffsetT;
        glLoadMatrixf( f );
      }
      glMatrixMode( GL_MODELVIEW );
    }
  static GWTexFrame const unity;

  float mScaleS;
  float mScaleT;
  float mOffsetS;
  float mOffsetT;
};

struct GWState;

//  The plain state sets up MODULATE, SRC_ALPHA/ONE_MINUS_SRC_ALPHA, white color 
//  but turns off lighting, blending and alpha testing.
extern GWState const GW_PLAIN_STATE;

struct GWState {
  GWState( GWState const & o = GW_PLAIN_STATE ) : 
    texture( o.texture ), texMode( o.texMode ), blendSrc( o.blendSrc ), blendDst( o.blendDst ), 
    color( o.color ), testLevel( o.testLevel ), texFrame( o.texFrame ),
    lit( o.lit ), blend( o.blend ), test( o.test )
    {
    }
  GWState( TextureRef const & tr, GLenum tm, GLenum bs, GLenum bd, GWColor c, float ml, 
      GWTexFrame const & tf, bool l, bool b, bool t, bool cl ) :
    texture( tr ), texMode( tm ), blendSrc( bs ), blendDst( bd ), 
    color( c ), testLevel( ml ), texFrame( tf ), 
    lit( l ), blend( b ), test( t )
    {
    }
  GWState const & operator=( GWState const & o )
    {
      texture = o.texture;
      texMode = o.texMode;
      blendSrc = o.blendSrc;
      blendDst = o.blendDst;
      color = o.color;
      testLevel = o.testLevel;
      texFrame = o.texFrame;
      lit = o.lit;
      blend = o.blend;
      test = o.test;
      return *this;
    }

  //  Pass NULL for arrays that you're not using. If you pass NULL for indices, "count" 
  //  means number of vertices in the array; else it means number of indices.
  //  This function assumes all arrays live separately, and that "pos" is an array of 
  //  3-vectors, "uv" is an array of 2-vectors, "norm" is an array of 3-vectors, and 
  //  color is an array of GWColor.
  void draw( GLenum primitive, unsigned int const * indices, unsigned int count, 
      float const * pos, float const * uv, float const * norm, GWColor const * color ) const;

  //todo: Factor out mode/combine/color into a more flexible block 
  //      and use a pointer to it. Provide defaults for REPLACE, 
  //      MODULATE, etc.
  //      Pehaps also for blending, alpha testing, etc.
  TextureRef        texture;
  GLenum            texMode;
  GLenum            blendSrc;
  GLenum            blendDst;
  GWColor           color;
  float             testLevel;
  GWTexFrame        texFrame;
  bool              lit;
  bool              blend;
  bool              test;
};

//  Call gwInit() to set up necessary internal GW state before 
//  starting to use GW calls. Call gwReset() after gwInit().
void gwInit();

//  gwSwapBuffers() will call glutSwapBuffers(), and in addition 
//  display the given overlay text on top of the frame.
void gwSwapBuffers();

//  gwAddOverlayText() adds some text (printf()-style) to be 
//  displayed on top of the image, once gwSwapBuffers() is called.
//  Once you've added text, it will be visible for several frames; 
//  you don't need to add the text again.
//  NOT CURRENTLY IMPLEMENTED
void gwAddOverlayText( char const * str, ... );

//  gwAddOverlayTextV() adds some text (vprintf()-style) to be 
//  displayed on top of the image, once gwSwapBuffers() is called.
//  Once you've added text, it will be visible for several frames; 
//  you don't need to add the text again.
//  NOT CURRENTLY IMPLEMENTED
void gwAddOverlayTextV( char const * str, va_list args );

//  gwForceState() will force the GL state to conform to the 
//  state described. It will also turn off any internally set 
//  state such as glVertexPointer()/enables.
void gwForceState( GWState const * state );

//  gwFlushState() tells GW that you've changed GL state, and 
//  would it please go ahead and assume everything is not 
//  what GW last thought it would be. This also includes the 
//  vertex/normal/uv/color pointers/enables.
void gwFlushState();

//  gwSetState() is the work-horse of GW, changing GL state 
//  to be what "state" describes. This is done using cached 
//  state, so that only necessary state change requests are 
//  issued.
void gwSetState( GWState const * state );

//  gwGetState() returns what GW thinks is the current GL state.
void gwGetState( GWState * outState );

//  gwReset() will force all GL state to plain vanilla defaults, 
//  including the various matrices. It is a superset of gwForceState() 
//  and gwFlushState().
void gwReset();

//  gwTextureCoordTransform() is a convenience wrapper for setting 
//  up the GL transform matrix to display some sub-piece of a texture.
void gwTextureCoordTransform( float offsetS, float offsetT, float scaleS, float scaleT );


#endif  //  glwrap_h
